/***************************************************************************
 *   Copyright (C) 2005 by Christian Muehlhaeuser, Last.fm Ltd.            *
 *   chris@last.fm                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include <QHttp>
#include <QUrl>
#include <QMutex>
#include <QTcpSocket>
#include <QTcpServer>
#include <QMessageBox>

#include <unistd.h>

#include "rtaudio/RtAudio.h"
#include "libsamplerate/samplerate.h"

#include "playback.h"
#include "settings.h"
#include "webserviceconnector.h"

#define BUFFER 16384
#define PLAYBACKBUFFER 4096

RtAudio *audio;
char *buffer;
QByteArray *out = new QByteArray();

QHttp *http;
QByteArray *waveBuffer;
QMutex mutex;
struct mpstr_tag mpeg;

SRC_STATE *src_state;
SRC_DATA src_data;

int volume = 1;
bool playbackStopped = false;

Playback *Playback::s_instance = 0;


int audioCallback( char *buffer, int bufferSize, void *data )
{
    int bufs = bufferSize * 4;
    memset( buffer, 0, bufs );

    if ( out->size() > bufs && !playbackStopped )
    {
        mutex.lock();
        memcpy( buffer, out->data(), bufs );
        out->remove( 0, bufs );
        mutex.unlock();
    }

    return 0;
}


void
WaveThread::run()
{
    char outBuf[16384];
    int size;

    while ( !playbackStopped )
    {
        mutex.lock();
        out->reserve( PLAYBACKBUFFER * 10 );

        if ( waveBuffer->size() >= PLAYBACKBUFFER && out->size() < 32768 )
        {
            QByteArray *in = new QByteArray( waveBuffer->left( PLAYBACKBUFFER ) );
            waveBuffer->remove( 0, PLAYBACKBUFFER );
            mutex.unlock();

            int result = decodeMP3( &mpeg, (unsigned char*)in->data(), PLAYBACKBUFFER, outBuf, sizeof( outBuf ), &size );

            while ( result == MP3_OK )
            {
                if ( src_data.src_ratio != 1.0 )
                {
                    float input[size];
                    float output[size];

                    memset( input, 0, sizeof( input ) );
                    memset( output, 0, sizeof( output ) );

                    src_data.end_of_input = 0;
                    src_data.input_frames = size / 4;
                    src_data.output_frames = size / 2;
                    src_data.data_in = input;
                    src_data.data_out = output;

                    src_short_to_float_array( (short *)outBuf, input, size / 2 );
                    memset( outBuf, 0, sizeof( outBuf ) );

                    src_process( src_state, &src_data );
                    src_float_to_short_array( output, (short *)outBuf, src_data.output_frames_gen * 2 );

                    size = src_data.output_frames_gen * 4;
                }

                float scale = (float)volume / 100;

                mutex.lock();
                // Append it the binary-safe way!
                for ( int i = 0; i < ( size / 2 ); i++ )
                {
                    union PCMDATA
                    {
                        short i;
                        unsigned char b[2];
                    } pcmData;

                    pcmData.b[0] = outBuf[i * 2];
                    pcmData.b[1] = outBuf[i * 2 + 1];

                    float pcmValue = (float)pcmData.i * scale;
                    pcmData.i = (short)pcmValue;

                    out->append( pcmData.b[0] );
                    out->append( pcmData.b[1] );
                }
                mutex.unlock();

                result = decodeMP3( &mpeg, NULL, 0, outBuf, sizeof( outBuf ), &size );
            }

            delete in;

        } else
        {
            mutex.unlock();
            this->msleep( 10 );
        }
    }

    playbackStopped = false;
    exit();
}


Playback::Playback()
{
    audio = 0;
    s_instance = this;
    playing = false;

    InitMP3( &mpeg );
    waveBuffer = new QByteArray();
    waveBuffer->reserve( BUFFER * 2 );

    int externalSystem = -1;
    #ifdef WIN32
    externalSystem = 1;
    #endif
    #ifdef Q_WS_X11
    externalSystem = 2;
    #endif

    if ( Settings::instance()->soundSystem() == externalSystem )
        Settings::instance()->setActAsProxy( true );
    else
        Settings::instance()->setActAsProxy( false );

    http = new QHttp( this );
    connect( http, SIGNAL( readyRead( QHttpResponseHeader ) ), this, SLOT( dataAvailable( QHttpResponseHeader ) ) );
    connect( http, SIGNAL( responseHeaderReceived( QHttpResponseHeader ) ), this, SLOT( responseHeaderReceived( QHttpResponseHeader ) ) );
    connect( http, SIGNAL( stateChanged( int ) ), this, SLOT( stateChanged( int ) ) );

    if ( Settings::instance()->actAsProxy() )
    {
        proxyServerSocket = new QTcpServer( this );
        connect( proxyServerSocket, SIGNAL( newConnection() ), this, SLOT( proxyClientConnect() ) );

        proxyServerSocket->listen( QHostAddress( "127.0.0.1" ), 32214 );
    }

    waveThread = new WaveThread();
}


Playback::~Playback()
{
    stopPlayback();
    qDebug() << "Deinit playback";

    if ( Settings::instance()->actAsProxy() )
    {
        proxyServerSocket->close();
        delete proxyServerSocket;
        proxyServerSocket = 0;
    }

    delete http;
    delete waveBuffer;
}


bool
Playback::startPlayback( QUrl url )
{
    if ( !initSound() )
        return false;

    mutex.lock();
    if ( waveThread->isRunning() )
    {
        playbackStopped = true;
        waveThread->wait( 100 );
    }

    waveBuffer->truncate( 0 );
    mutex.unlock();

    if ( !Settings::instance()->actAsProxy() )
        audio->startStream();

    if ( Settings::instance()->proxyUsage() && !Settings::instance()->proxyHost().isEmpty() )
        http->setProxy( Settings::instance()->proxyHost(), Settings::instance()->proxyPort(),
                        Settings::instance()->proxyUsername(), Settings::instance()->proxyPassword() );

    http->setHost( url.host(), url.port() > 0 ? url.port() : 80 );
    if ( !url.encodedQuery().isEmpty() )
        http->get( url.path() + "?" + QString( url.encodedQuery() ) );
    else
        http->get( url.path() );

    return true;
}


void
Playback::stopPlayback()
{
    if ( !isPlaying() )
        return;

    playing = false;
    qDebug( "Stopping playback!" );

    mutex.lock();
    http->abort();

    if ( waveThread->isRunning() )
    {
        playbackStopped = true;
        waveThread->wait( 100 );
    }

    waveBuffer->truncate( 0 );
    out->truncate( 0 );
    mutex.unlock();

    src_delete( src_state );

    qDebug() << "Deinit sockets";
    if ( Settings::instance()->actAsProxy() )
    {
        if ( proxySocketList.count() > 0 )
        {
            for ( int i = 0; i < proxySocketList.count(); i++ )
            {
                if ( proxySocketList.at( i )->state() == QAbstractSocket::ConnectedState )
                {
                    proxySocketList.at( i )->flush();
                    proxySocketList.at( i )->close();
                }

                delete proxySocketList.at( i );
            }
        }

        proxySocketList.clear();
        qDebug() << "Proxy clients left:" << proxySocketList.count();
    }
    else
    {
        if ( audio != 0 )
        {
            audio->stopStream();
            audio->closeStream();

            delete audio;
            audio = 0;
        }
    }
}


bool
Playback::isPlaying()
{
    return playing;
}


void
Playback::setVolume( int percentage )
{
    volume = percentage + 1;
}


void
Playback::setDirectSkip()
{
    m_directSkip = true;
}


bool
Playback::initSound()
{
    if ( Settings::instance()->actAsProxy() )
        return true;

    char data[2];
    int channels = 2;
    int nBuffers = 8;
    int bufferSize = 512;
    int sampleRate = -1;

    try
    {
        RtAudio::RtAudioApi api = RtAudio::UNSPECIFIED;
        RtAudioFormat format = RTAUDIO_SINT16;

        #ifdef Q_WS_X11
        switch ( Settings::instance()->soundSystem() )
        {
            case 0:
                qDebug( "Using ALSA!" );
                api = RtAudio::LINUX_ALSA;
                break;

            case 1:
                qDebug( "Using OSS!" );
                api = RtAudio::LINUX_OSS;
                break;
        }
        #endif

        audio = new RtAudio();
        RtAudioDeviceInfo info = audio->getDeviceInfo( Settings::instance()->soundCard() + 1 );
        if ( info.nativeFormats & RTAUDIO_SINT32 )
        {
            qDebug() << "Supports 32 bit";
            format = RTAUDIO_SINT32;
        }
        if ( info.nativeFormats & RTAUDIO_SINT24 )
        {
            qDebug() << "Supports 24 bit";
            format = RTAUDIO_SINT24;
        }
        if ( info.nativeFormats & RTAUDIO_SINT16 )
        {
            qDebug() << "Supports 16 bit";
            format = RTAUDIO_SINT16;
        }

        for ( uint j = 0; j < info.sampleRates.size(); j++ )
        {
            if ( abs( info.sampleRates.at( j ) - 44100 ) < abs( sampleRate - 44100 ) )
                sampleRate = info.sampleRates.at( j );
        }

        delete audio;
        audio = new RtAudio( Settings::instance()->soundCard(), channels, 0, 0, format,
                             sampleRate, &bufferSize, nBuffers, api );

        int error;
        src_state = src_new( SRC_SINC_MEDIUM_QUALITY, 2, &error );

        src_data.end_of_input = 0;
        src_data.input_frames = 0;
        src_data.output_frames = 0;
        src_data.src_ratio = (double)sampleRate / 44100.0;
    }
    catch ( RtError &error )
    {
        error.printMessage();
        delete audio;
        audio = 0;

        return false;
    }

    audio->setStreamCallback( &audioCallback, (void *)data );
    return true;
}


void
Playback::dataAvailable( const QHttpResponseHeader &resp )
{
    if ( !m_directSkip && ( waveBuffer->size() >= BUFFER * 2 ) )
        return;

    memset( inBuf, 0, sizeof( inBuf ) );
    int len = http->read( inBuf, sizeof( inBuf ) );

    if ( resp.statusCode() == 401 || strstr( inBuf, "HTTP/1.0 401" ) != NULL )
    {
        showErrorCode( 401 );
        return;
    }
    if ( resp.statusCode() == 503 || strstr( inBuf, "HTTP/1.0 503" ) != NULL )
    {
        showErrorCode( 503 );
        return;
    }
    if ( resp.statusCode() == 666 || strstr( inBuf, "HTTP/1.0 666" ) != NULL )
    {
        showErrorCode( 666 );
        return;
    }
    if ( resp.statusCode() == 667 || strstr( inBuf, "HTTP/1.0 667" ) != NULL )
    {
        showErrorCode( 667 );
        return;
    }

    mutex.lock();

    // Append it the binary-safe way!
    for ( int i = 0; i < len; i++ )
        waveBuffer->append( inBuf[ i ] );

    if ( waveBuffer->indexOf( "SYNC" ) >= 0 )
    {
        qDebug( "New song detected!" );
        WebserviceConnector::instance()->metaData();

        if ( m_directSkip )
        {
            m_directSkip = false;
            if ( waveBuffer->size() > BUFFER )
                waveBuffer->remove( BUFFER, waveBuffer->size() );
        }

        waveBuffer->remove( waveBuffer->indexOf( "SYNC" ), 4 );
    }

    if ( Settings::instance()->actAsProxy() )
    {
        if ( proxySocketList.count() > 0 )
            for ( int i = 0; i < proxySocketList.count(); i++ )
            {
                if ( proxySocketList.at( i )->state() == QAbstractSocket::ConnectedState )
                    proxySocketList.at( i )->write( waveBuffer->data(), waveBuffer->size() );
            }

        waveBuffer->truncate( 0 );
    }

    if ( ( ( waveBuffer->size() + len ) < PLAYBACKBUFFER ) && waveThread->isRunning() )
    {
        qDebug( "Buffer empty!" );
        playbackStopped = true;
    }

    if ( !waveThread->isRunning() && waveBuffer->size() >= BUFFER )
        waveThread->start( QThread::TimeCriticalPriority );

    mutex.unlock();
}


void
Playback::stateChanged( int state )
{
    if ( !playing && state == QHttp::Reading )
    {
        playing = true;
        emit playbackStarting();
    }

    if ( state == QHttp::Unconnected )
    {
        stopPlayback();
        emit playbackFinished();
    }
}


void
Playback::responseHeaderReceived( const QHttpResponseHeader &resp )
{
    if ( resp.statusCode() == 503 || resp.statusCode() == 401 || resp.statusCode() == 666 )
        showErrorCode( resp.statusCode() );
}


void
Playback::showErrorCode( int error )
{
    qDebug( "Streamer disconnected us!" );

    switch( error )
    {
        case 401:
            QMessageBox::information( 0, "Attention", "Your session is invalid! Please login, again!\n" );
            break;

        case 503:
            QMessageBox::information( 0, "Attention", "No more streaming slots available. Try again in a few minutes!\n" );
            break;

        case 666:
            QMessageBox::information( 0, "Attention", "Streamer is shutting down for maintenance! Try again in a few minutes!\n" );
            break;

        case 667:
            QMessageBox::information( 0, "Error", "There is not enough content left to play this station.\n" );
            break;
    }
}


void
Playback::proxyClientConnect()
{
    qDebug() << "New proxy client connected!";

    QTcpSocket *socket = proxyServerSocket->nextPendingConnection();
    connect( socket, SIGNAL( disconnected() ), this, SLOT( proxyClientDisconnect() ) );

    // welcome the new client
    QByteArray token( "HTTP/1.1 200 OK\n\n" );
    socket->write( token, token.length() );

    proxySocketList.append( socket );
    if ( proxySocketList.count() == 1 )
        emit proxyClientConnected();

    qDebug() << "Proxy clients atm:" << proxySocketList.count();
}


void
Playback::proxyClientDisconnect()
{
    qDebug() << "Proxy client disconnected!";

    #ifdef WIN32
    Sleep( 500 );
    #endif
    #ifndef WIN32
    usleep( 500 );
    #endif

    QApplication::processEvents();

    for ( int i = 0; i < proxySocketList.count(); i++ )
    {
        if ( proxySocketList.at( i )->state() == QAbstractSocket::ConnectedState )
        {
            return;
        }
    }

    stopPlayback();
    emit proxyClientDisconnected();
}
